#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "global.h"

/* This file contains the binary tree and its functions, used for keeping
variables. */

/* Commented functions are meant to be used externally */


// Part of the binary tree keeping variables
struct node {
	struct variable *key;
	struct node *left;
	struct node *right;
};

struct node *createLeaf(const char s[], double val);
void deleteLeaf(struct node *leaf);
void insertUnder(const char s[], double val, struct node *leaf);
struct node *searchUnder(const char s[], struct node *leaf);
int printVarLeaf(const struct node *leaf);

struct node *root = NULL; // Root of binary tree

/* Removes all variables in the tree */
void clearTree(void)
{
	deleteLeaf(root);
	root = NULL;
}

struct node *createLeaf(const char s[], double val)
{
	struct node *leaf;
	if ((leaf = malloc(sizeof(*leaf))) == NULL) {
		exit(EXIT_FAILURE);
	}
	if ((leaf->key = malloc(sizeof(*leaf->key))) == NULL) {
		exit(EXIT_FAILURE);
	}
	if ((leaf->key->name = malloc(strlen(s) * sizeof(*s))) == NULL) {
		exit(EXIT_FAILURE);
	}
	strcpy(leaf->key->name, s);
	leaf->key->value = val;
	leaf->left = NULL;
	leaf->right = NULL;
	return leaf;
}

void deleteLeaf(struct node *leaf)
{
	if (leaf != NULL) {
		deleteLeaf(leaf->left);
		deleteLeaf(leaf->right);
		free(leaf->key->name);
		free(leaf->key);
		free(leaf);
	}
}

/* inserts variable with name s and value val into the tree */
void insert(const char s[], double val)
{
	if (root != NULL) {
		insertUnder(s, val, root);
	}
	else {
		root = createLeaf(s, val);
	}
}

void insertUnder(const char s[], double val, struct node *leaf)
{
	if (strcmp(s, leaf->key->name) < 0) {
		if (leaf->left != NULL) {
			insertUnder(s, val, leaf->left);
		}
		else {
			leaf->left = createLeaf(s, val);
		}
	}
	else {
		if (leaf->right != NULL) {
			insertUnder(s, val, leaf->right);
		}
		else {
			leaf->right = createLeaf(s, val);
		}
	}
}

/* Searches for variable with name s and returns it. Returns NULL if it does
not exist. */
struct variable *search(const char s[])
{
	struct node *r;

	r = searchUnder(s, root);
	if (r != NULL) {
		return r->key;
	}
	return NULL;
}

struct node *searchUnder(const char s[], struct node *leaf)
{
	if (leaf != NULL) {
		if (!strcmp(s, leaf->key->name)) {
			return leaf;
		}
		else if (strcmp(s, leaf->key->name) < 0) {
			return searchUnder(s, leaf->left);
		}
		else {
			return searchUnder(s, leaf->right);
		}
	}
	return NULL;
}

/* Prints all variables in tree with their respective values. */
void printVars(void)
{
	if (printVarLeaf(root) == 0) {
		printf("\tNo variables in memory.\n");
	}
}

int printVarLeaf(const struct node *leaf)
{
	int len = 0;

	if (leaf != NULL) {
		len += printVarLeaf(leaf->left);
		printf("\t%s : %.8g\n", leaf->key->name, leaf->key->value);
		len++;
		len += printVarLeaf(leaf->right);
	}
	return len;
}
