#include <stdio.h>
#include <ctype.h>

#include "global.h"

#define BUFSIZE 100

void getIdentifier(char s[], int i, int maxlen);
int getch(void);
void ungetch(int);

/* getop: get next operator, identifier or numeric operand */
int getop(char s[], int maxlen)
{
	int i, c;

	while ((s[0] = c = getch()) == ' ' || c == '\t')
		;
	s[1] = '\0';

	if (!isdigit(c) && !isalpha(c) && c != '.' && c != '-') {
		return c;	/* not a number or an identifier */
	}

	i = 0;
	if (isalpha(c)) /* collect letter sequence */
	{
		getIdentifier(s, i, maxlen);
		return IDENTIFIER;
	}
	if (c == '-')
	{
		if (isdigit(c = getch()) || c == '.')
			s[++i] = c;
		else
		{
			if (c != EOF)
				ungetch(c);
			return '-';
		}
	}

	if (isdigit(c))	/* collect integer part */
		while (isdigit(s[++i] = c = getch())) {
			if (i == maxlen - 1) {
				printf("error: Number is too long,");
				printf(" Digits after digit number %d ignored.\n", maxlen - 2);
				while (isdigit((c) = getch()))
					;
				break;
			}
		}
	if (c == '.')	/* collect fraction part */
		while (isdigit(s[++i] = c = getch())) {
			if (i == maxlen - 3) {
				printf("error: Number is too long,");
				printf(" Digits after digit number %d ignored.\n", maxlen - 3);
				while (isdigit((c) = getch()))
					;
				break;
			}
		}
	s[i] = '\0';
	if (c != EOF)
		ungetch(c);
	return NUMBER;
}

char buf[BUFSIZE];	/* buffer for ungetch */
int bufp = 0;		/* next free position in buf */

int getch(void)	/* get a (possibly pushed back) character */
{
	return (bufp > 0) ? buf[--bufp] : getchar();
}

void ungetch(int c)	/* push character back on input */
{
	if (bufp >= BUFSIZE) {
		printf("ungetch: too many characters\n");
	}
	else {
		buf[bufp++] = c;
	}
}

/* get sequence of letters, numbers and "_"s from stdin,
and save them in string s starting at index position i */
void getIdentifier(char s[], int i, int maxlen)
{
	char c;

	while (isalpha(s[++i] = c = getch()) || isdigit(s[i]) || s[i] == '_') {
		if (i == maxlen - 2) {
			printf("error: Variable name is too long,");
			printf(" characters after character %d ignored.\n", maxlen - 2);
				while (isalpha((c) = getch()))
					;
			break;
		}
	}
	s[i] = '\0';

	if (c != EOF)
		ungetch(c);
}
