#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "global.h"

#define MAXOP	100		/* max size of an operand or operator */

// external
int getop(char s[], int maxlen);
void push(double f);
double pop(void);
double getvalue(void);
double getvalue2(void);
void duplicate(void);
void swap(void);
void clear(void);
int getStackSize(void);

// external, for variables
void clearTree(void);
void insert(const char s[], double val);
struct variable *search(const char s[]);
void printVars(void);

// local
void handleIdentifiers(const char s[]);
void handleVars(const char s[]);
void assignVar(const char s[]);
void power(void);
void squareroot(void);

char *lastvar = NULL; /* Last variable or NULL if previous element was not 
						 variable */
double ans = 0;	/* answer to previous calculation */

/* reverse Polish calculator */
int main()
{
	int type;
	char s[MAXOP];

	while ((type = getop(s, MAXOP)) != EOF)
	{
		if (type != '=' && lastvar != NULL) {
			free(lastvar);
			lastvar = NULL;
		}
		switch (type) {
			case NUMBER:
				push(atof(s));
				break;
			case IDENTIFIER:
				handleIdentifiers(s);
				break;
			case '=':
				assignVar(lastvar);
				break;
			case '+':
				push(pop() + pop());
				break;
			case '*':
				push(pop() * pop());
				break;
			case '-':
				swap();
				push(pop() - pop());
				break;
			case '/':
				if (getvalue() != 0.0)
				{
					swap();
					push(pop() / pop());
				}
				else
					printf("error: zero divisor\n");
				break;
			case '%':
				if (getvalue() != 0.0)
				{
					push(fmod(pop(), pop()));
				}
				else
					printf("error: zero divisor\n");
				break;
			case '~':
				swap();
				break;
			case '#':
				duplicate();
				break;
			case '^':
				power();
				break;
			case '\n':
				if (getStackSize() != 0) {
					ans = pop();
				}
				else {
					ans = 0;
				}
				printf("\t%.8g\n", ans);
				clear();
				push(ans);
				break;
			default:
				printf("error: unknown command %s\n", s);
				break;
		}
	}
	return EXIT_SUCCESS;
}

void handleIdentifiers(const char s[])
{
	if (!strcmp(s, "pow")) {
		power();
	}
	else if (!strcmp(s, "exp")) {
		push(exp(pop()));
	}
	else if(!strcmp(s, "sin")) {
		push(sin(pop()));
	}
	else if(!strcmp(s, "sqrt")) {
		squareroot();
	}
	else if(!strcmp(s, "pi")) {
		push(M_PI);
	}
	else if(!strcmp(s, "e")) {
		push(M_E);
	}
	else if(!strcmp(s, "ans")) {
		push(ans);
	}
	else if(!strcmp(s, "clearvars") || !strcmp(s, "cv")) {
		clearTree();
	}
	else if(!strcmp(s, "printvars") || !strcmp(s, "pv")) {
		printVars();
	}
	else {
		handleVars(s);
	}
}

void handleVars(const char s[])
{
	struct variable *var;

	var = search(s);
	if (var != NULL) {
		push(var->value);
	}
	else {
		push(0);
	}

	if ((lastvar = malloc(strlen(s) * sizeof(*s))) == NULL) {
		exit(EXIT_FAILURE);
	}
	strcpy(lastvar, s);
}

void assignVar(const char s[])
{
	if (s != NULL)
	{
		struct variable *var;

		var = search(s);
		pop(); // removing value of variable from stack
		if (var != NULL) {
			var->value = getvalue();
		}
		else {
			insert(s, getvalue());
		}
	}
	else
		printf("error: element before = was not a valid variable.\n");
}

void power(void)
{
	if (getvalue() != 0.0 || getvalue2() != 0.0)
	{
		if (getvalue2() >= 0 || getvalue() == round(getvalue()))
			push(pow(pop(), pop()));
		else
		{
			printf("error: x cannot be negative if y is");
			printf(" fractional.\n");
		}
	}
	else
		printf("error: zero divisor\n");
}

void squareroot(void)
{
	if (getvalue() < 0.0) {
		printf("error: no square root of %.8g\n", getvalue());
	}
	else {
		push(sqrt(pop()));
	}
}
